
// see: https://www.rfc-editor.org/rfc/rfc6901#section-3
export function escape_for_json_pointer(str: string, escape_for_markdown = true) {
	return str
		.replace(/~/g, escape_for_markdown ? '\\~0' : '~0')
		.replace(/\//g, escape_for_markdown ? '\\~1' : '~1');
}

export function unescape_for_json_pointer(str: string) {
	return str.replace(/~1/g, '/').replace(/~0/g, '~');
}

export interface ResolvedPointer {
	value: any;
	found: boolean;
	stack: {
		parent: any;
		field: string;
	}[];
}

export function resolve_json_pointer(root: object, json_pointer: string) : ResolvedPointer {
	if (json_pointer === '' || json_pointer === '/') {
		return {
			value: root,
			found: true,
			stack: [ ],
		};
	}

	if (! json_pointer.startsWith('/')) {
		throw new Error('invalid JSON pointer');
	}

	if (json_pointer.includes('//')) {
		throw new Error('invalid JSON pointer');
	}

	const fields = json_pointer.split('/').slice(1);

	if (json_pointer.endsWith('/')) {
		fields.pop();
	}

	const resolved: ResolvedPointer = {
		value: null,
		found: true,
		stack: [ ],
	};

	let current = root;
	
	for (let field of fields) {
		field = unescape_for_json_pointer(field);
		resolved.stack.push({ parent: current, field });

		if (! (field in current)) {
			resolved.found = false;
			break;
		}

		current = current[field];
	}

	if (resolved.found) {
		resolved.value = current;
	}

	return resolved;
}

export function jsonptr(...steps: string[]) {
	return new JsonPointer(steps);
}

export function jsonptr_str(...steps: string[]) {
	return (new JsonPointer(steps)).as_str();
}

export function jsonptr_md_str(...steps: string[]) {
	return (new JsonPointer(steps)).as_md_str();
}

export class JsonPointer {
	public steps: string[];
	public md_steps: string[];

	private _str: string;
	private _md_str: string;

	constructor(steps: string[] = [ ]) {
		this.steps = steps.map((step) => escape_for_json_pointer(step, false));
		this.md_steps = steps.map((step) => escape_for_json_pointer(step, true));
	}

	public as_str() {
		return this._str = (this._str || '/' + this.steps.join('/'));
	}

	public as_md_str() {
		return this._md_str = (this._md_str || '/' + this.md_steps.join('/'));
	}

	public step_down(next: string) {
		const new_pointer = new JsonPointer();
		new_pointer.steps.push(...this.steps, escape_for_json_pointer(next, false));
		new_pointer.md_steps.push(...this.md_steps, escape_for_json_pointer(next, true));
		return new_pointer;
	}
}
